@@ -1,49 +1,21 @@
 /*
+ * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
+ * @license AGPL-3.0
  *
- *    OPEN-XCHANGE legal information
+ * This code is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU Affero General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
- *    All intellectual property rights in the Software are protected by
- *    international copyright laws.
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU Affero General Public License for more details.
  *
+ * You should have received a copy of the GNU Affero General Public License
+ * along with OX App Suite.  If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
  *
- *    In some countries OX, OX Open-Xchange, open xchange and OXtender
- *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the OX Software GmbH group of companies.
- *    The use of the Logos is not covered by the GNU General Public License.
- *    Instead, you are allowed to use these Logos according to the terms and
- *    conditions of the Creative Commons License, Version 2.5, Attribution,
- *    Non-commercial, ShareAlike, and the interpretation of the term
- *    Non-commercial applicable to the aforementioned license is published
- *    on the web site http://www.open-xchange.com/EN/legal/index.html.
- *
- *    Please make sure that third-party modules and libraries are used
- *    according to their respective licenses.
- *
- *    Any modifications to this package must retain all copyright notices
- *    of the original copyright holder(s) for the original code used.
- *
- *    After any such modifications, the original and derivative code shall remain
- *    under the copyright of the copyright holder(s) and/or original author(s)per
- *    the Attribution and Assignment Agreement that can be located at
- *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
- *    given Attribution for the derivative code and a license granting use.
- *
- *     Copyright (C) 2016-2020 OX Software GmbH
- *     Mail: info@open-xchange.com
- *
- *
- *     This program is free software; you can redistribute it and/or modify it
- *     under the terms of the GNU General Public License, Version 2 as published
- *     by the Free Software Foundation.
- *
- *     This program is distributed in the hope that it will be useful, but
- *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
- *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
- *     for more details.
- *
- *     You should have received a copy of the GNU General Public License along
- *     with this program; if not, write to the Free Software Foundation, Inc., 59
- *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
+ * Any use of the work other than as authorized under this license or copyright law is prohibited.
  *
  */
 
@@ -67,17 +39,18 @@ import javax.mail.Message;
 import javax.mail.MessagingException;
 import javax.mail.UIDFolder;
 import javax.mail.internet.InternetAddress;
+import javax.mail.internet.MimeMessage.RecipientType;
+import javax.mail.search.FlagTerm;
 import org.slf4j.Logger;
 import com.openexchange.context.ContextService;
 import com.openexchange.exception.OXException;
 import com.openexchange.groupware.contexts.Context;
+import com.openexchange.imap.IMAPAccess;
 import com.openexchange.imap.IMAPCapabilities;
-import com.openexchange.imap.IMAPFolderStorage;
 import com.openexchange.imap.IMAPProvider;
 import com.openexchange.java.Strings;
 import com.openexchange.mail.Protocol;
 import com.openexchange.mail.api.IMailFolderStorage;
-import com.openexchange.mail.api.IMailFolderStorageDelegator;
 import com.openexchange.mail.api.IMailMessageStorage;
 import com.openexchange.mail.api.MailAccess;
 import com.openexchange.mail.api.MailCapabilities;
@@ -368,8 +341,13 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         return notifications;
     }
 
+    /** Simple class to delay initialization until needed */
+    private static class RecentTermHolder {
+        static final FlagTerm TERM_RECENT = new FlagTerm(new Flags(Flags.Flag.RECENT), true);
+    }
+
     @Override
-    public void run() {
+    public synchronized void run() {
         if (canceled.get()) {
             return;
         }
@@ -402,8 +380,9 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                 mailAccess.connect(false);
 
                 boolean notified = false;
+                final MailAccess<? extends IMailFolderStorage, ? extends IMailMessageStorage> mailAccess1 = mailAccess;
 
-                IMAPStore imapStore = getImapFolderStorageFrom(mailAccess).getImapStore();
+                IMAPStore imapStore = IMAPAccess.getIMAPFolderStorageFrom(mailAccess1).getImapStore();
                 final IMAPFolder imapFolder = (IMAPFolder) imapStore.getFolder(fullName);
                 this.imapFolderInUse = imapFolder;
                 Map<String, Object> props = new HashMap<String, Object>(3);
@@ -438,9 +417,15 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
                     {
                         int newMessageCount = imapFolder.getNewMessageCount();
                         if (newMessageCount > 0) {
-                            LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
-                            notifyNewMail();
-                            notified = true;
+                            // Search for recent message
+                            Message[] recentMessages = imapFolder.search(RecentTermHolder.TERM_RECENT);
+                            if (recentMessages != null && recentMessages.length > 0) {
+                                LOGGER.debug("IMAP-IDLE result for user {} in context {}: Doing push due to {} new mail(s)", sUserId, sContextId, Integer.toString(newMessageCount));
+                                setEventPropertiesByRecent(recentMessages, props, imapFolder);
+                                notifyNewMail();
+                                LOGGER.debug("Performed IMAP-IDLE run having new messages for user {} in context {}. ", sUserId, sContextId);
+                                return;
+                            }
                         }
                     }
 
@@ -787,6 +772,31 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         }
     }
 
+    private void setEventPropertiesByRecent(Message[] recentMessages, Map<String, Object> props, IMAPFolder imapFolder) {
+        try {
+            imapFolder.fetch(recentMessages, FETCH_PROFILE_MSG_INFO);
+
+            StringBuilder buf = new StringBuilder(64);
+            long[] uids = new long[recentMessages.length];
+            for (int i = recentMessages.length; i-- > 0;) {
+                long uid = ((IMAPMessage) recentMessages[i]).getUID();
+                if (buf.length() > 0) {
+                    buf.append(',');
+                }
+                buf.append(uid);
+                uids[i] = uid;
+            }
+            props.put(PushEventConstants.PROPERTY_IDS, buf.toString());
+
+            Container<MailMessage> container = buildContainerFor(uids, recentMessages, imapFolder);
+            if (null != container) {
+                props.put(PushEventConstants.PROPERTY_CONTAINER, container);
+            }
+        } catch (MessagingException e) {
+            LOGGER.warn("Could not determine new message count.", e);
+        }
+    }
+
     private void setEventProperties(long uidNext, int totalCount, Map<String, Object> props, IMAPFolder imapFolder) {
         if (false == supportsPermanentListeners) {
             return;
@@ -846,30 +856,9 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
     private Container<MailMessage> fetchMessageInfoFor(long[] uids, IMAPFolder imapFolder) throws MessagingException {
         try {
-            int unread = imapFolder.getUnreadMessageCount();
             Message[] messages = imapFolder.getMessagesByUID(uids);
             imapFolder.fetch(messages, FETCH_PROFILE_MSG_INFO);
-
-            String fullName = imapFolder.getFullName();
-            Map<Long, MailMessage> map = new HashMap<Long, MailMessage>(messages.length);
-            for (Message message : messages) {
-                try {
-                    IMAPMessage im = (IMAPMessage) message;
-                    MailMessage mailMessage = convertMessage(im, fullName, unread);
-                    map.put(Long.valueOf(im.getUID()), mailMessage);
-                } catch (Exception e) {
-                    LOGGER.warn("Could not handle message.", e);
-                }
-            }
-
-            Container<MailMessage> container = new Container<MailMessage>();
-            for (long uid : uids) {
-                MailMessage mailMessage = map.get(Long.valueOf(uid));
-                if (null != mailMessage) {
-                    container.add(mailMessage);
-                }
-            }
-            return container;
+            return buildContainerFor(uids, messages, imapFolder);
         } catch (javax.mail.StoreClosedException e) {
             // Re-throw...
             throw e;
@@ -879,11 +868,57 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         }
     }
 
+    private Container<MailMessage> buildContainerFor(long[] uids, Message[] messages, IMAPFolder imapFolder) throws MessagingException {
+        int unread = imapFolder.getUnreadMessageCount();
+        String fullName = imapFolder.getFullName();
+        Map<Long, MailMessage> map = new HashMap<Long, MailMessage>(messages.length);
+        for (Message message : messages) {
+            try {
+                IMAPMessage im = (IMAPMessage) message;
+                MailMessage mailMessage = convertMessage(im, fullName, unread);
+                map.put(Long.valueOf(im.getUID()), mailMessage);
+            } catch (Exception e) {
+                LOGGER.warn("Could not handle message.", e);
+            }
+        }
+
+        Container<MailMessage> container = new Container<MailMessage>();
+        for (long uid : uids) {
+            MailMessage mailMessage = map.get(Long.valueOf(uid));
+            if (null != mailMessage) {
+                container.add(mailMessage);
+            }
+        }
+        return container;
+    }
+
     private MailMessage convertMessage(IMAPMessage im, String fullName, int unread) throws MessagingException, OXException {
         MailMessage mailMessage = new IDMailMessage(Long.toString(im.getUID()), fullName);
-        mailMessage.addFrom(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_FROM, im));
-        mailMessage.addTo(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_TO, im));
-        mailMessage.addCc(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_CC, im));
+
+        try {
+            mailMessage.addFrom((InternetAddress[]) im.getFrom());
+        } catch (MessagingException x) {
+            LOGGER.debug("Failed to get \"From\" address header", x);
+            mailMessage.addFrom(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_FROM, im));
+        }
+        try {
+            mailMessage.addTo((InternetAddress[]) im.getRecipients(RecipientType.TO));
+        } catch (MessagingException x) {
+            LOGGER.debug("Failed to get \"To\" address header", x);
+            mailMessage.addTo(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_TO, im));
+        }
+        try {
+            mailMessage.addCc((InternetAddress[]) im.getRecipients(RecipientType.CC));
+        } catch (MessagingException x) {
+            LOGGER.debug("Failed to get \"Cc\" address header", x);
+            mailMessage.addCc(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_CC, im));
+        }
+        try {
+            mailMessage.addBcc((InternetAddress[]) im.getRecipients(RecipientType.BCC));
+        } catch (MessagingException x) {
+            LOGGER.debug("Failed to get \"Bcc\" address header", x);
+            mailMessage.addBcc(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_BCC, im));
+        }
 
         final Flags msgFlags = im.getFlags();
         int flags = 0;
@@ -911,7 +946,6 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
 
         mailMessage.setFlags(flags);
 
-        mailMessage.addBcc(MimeMessageUtility.getAddressHeader(MessageHeaders.HDR_BCC, im));
         {
             String[] tmp = im.getHeader(MessageHeaders.HDR_CONTENT_TYPE);
             if ((tmp != null) && (tmp.length > 0)) {
@@ -927,7 +961,12 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
             // Size unavailable
             mailMessage.setSize(-1);
         }
-        mailMessage.setSubject(MimeMessageUtility.getSubject(im), true);
+        try {
+            mailMessage.setSubject(im.getSubject(), true);
+        } catch (MessagingException x) {
+            LOGGER.debug("Failed to get subject", x);
+            mailMessage.setSubject(MimeMessageUtility.getSubject(im), true);
+        }
         mailMessage.setUnreadMessages(unread);
         return mailMessage;
     }
@@ -947,18 +986,4 @@ public final class ImapIdlePushListener implements PushListener, Runnable {
         }
     }
 
-    private static IMAPFolderStorage getImapFolderStorageFrom(final MailAccess<? extends IMailFolderStorage, ? extends IMailMessageStorage> mailAccess) throws OXException {
-        IMailFolderStorage fstore = mailAccess.getFolderStorage();
-        if (!(fstore instanceof IMAPFolderStorage)) {
-            if (!(fstore instanceof IMailFolderStorageDelegator)) {
-                throw PushExceptionCodes.UNEXPECTED_ERROR.create("Unknown MAL implementation: " + fstore.getClass().getName());
-            }
-            fstore = ((IMailFolderStorageDelegator) fstore).getDelegateFolderStorage();
-            if (!(fstore instanceof IMAPFolderStorage)) {
-                throw PushExceptionCodes.UNEXPECTED_ERROR.create("Unknown MAL implementation: " + fstore.getClass().getName());
-            }
-        }
-        return (IMAPFolderStorage) fstore;
-    }
-
 }
